import os
import sys
import logging
import contextlib
import yaml
from typing import Union, Any, List, Dict, Tuple, Optional, cast, Iterator, Iterable
from abc import ABC
import subprocess

import jinja2

logger: logging.Logger = logging.getLogger(__name__)


class Subscriptable(ABC):
    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def get(self, key, default=None):
        return self.__dict__.get(key, default)


class UserError(Exception):
    pass


def realpath(path: str) -> str:
    return os.path.expandvars(os.path.expanduser(path))


def run_command(
    command: str,
    cwd: str,
    stdin: Optional[bytes] = None,
    decode_stdout=True,
    wait=True,
) -> Optional[Union[str, bytes]]:
    popenargs = {
        'args': command,
        'shell': True,
        'stdin': subprocess.PIPE,
        'stdout': subprocess.PIPE,
        'stderr': subprocess.PIPE,
        'cwd': cwd,
    }

    process = subprocess.Popen(**popenargs)
    if not wait:
        return
    try:
        stdout, stderr = process.communicate(stdin)
    except:
        process.kill()
        raise
    retcode = process.poll()
    if retcode:
        try:
            print(stderr.decode())
        except UnicodeDecodeError:
            print(stderr)
        raise subprocess.CalledProcessError(
            retcode,
            process.args,
            output=stdout,
            stderr=stderr,
        )

    if decode_stdout:
        return stdout.decode()
    else:
        return stdout


def render_template(
    contents: str,
    variables: Dict[str, Any],
    directory: str = None,
) -> str:
    env = jinja2.Environment(loader=jinja2.FileSystemLoader(directory)) \
        if directory is not None else jinja2.Environment()
    template = env.from_string(contents)
    rendered = template.render(variables)
    return rendered


@contextlib.contextmanager
def saved_sys_properties() -> Iterator[None]:
    """Save various sys properties such as sys.path and sys.modules."""
    old_path = sys.path.copy()
    old_modules = sys.modules.copy()

    try:
        yield
    finally:
        sys.path = old_path
        for module in set(sys.modules).difference(old_modules):
            del sys.modules[module]


def read_header(path: str) -> Dict[Any, Any]:
    header_lines = []
    with open(path, 'r') as f:
        for line in f:
            if line == '#\n':
                header_lines.append('\n')
            if line.startswith('# '):
                header_lines.append(line[2:])
            else:
                break
    header = yaml.safe_load(''.join(header_lines))
    if not header:
        header = {}
    if not isinstance(header, dict):
        raise UserError(f"invalid YAML header for {path}")
    return header



import os
from dataclasses import dataclass, field
from typing import List, Optional
from enum import Enum
from logging import CRITICAL, WARNING, INFO

from powar.util import Subscriptable


class AppMode(Enum):
    INSTALL = 0
    LIST_PACKAGES = 1
    NEW_MODULE = 2
    INIT = 3


class AppLogLevel(Enum):
    NORMAL = 0
    VERBOSE = 1
    QUIET = 2

    def into_logging_level(self) -> int:
        if self == AppLogLevel.NORMAL:
            return WARNING
        if self == AppLogLevel.VERBOSE:
            return INFO
        if self == AppLogLevel.QUIET:
            return CRITICAL
        assert False


@dataclass
class AppSettings(Subscriptable):
    data_path: str = os.path.join(os.path.dirname(__file__), "../data")
    module_config_template_filename: str = "module-template.py"
    global_config_template_filename: str = "global-template.py"

    template_dir: str = os.path.join(
        os.environ.get("XDG_CONFIG_HOME", "$HOME/.config"), "powar-templates")

    config_dir: str = os.path.join(
        os.environ.get("XDG_CONFIG_HOME", "$HOME/.config"), "powar-config")

    cache_dir: str = os.path.join(
        os.environ.get("XDG_DATA_HOME", "$HOME/.local/share"), "powar")

    module_config_filename: str = "powar.py"
    global_config_filename: str = "global.py"
    modules_to_consider: List[str] = field(default_factory=list)

    log_level: AppLogLevel = AppLogLevel.NORMAL

    dry_run: bool = False

    mode: Optional[AppMode] = None

    new_module_name: Optional[str] = None

    init: bool = False

    switch_to_root: bool = False



import logging
import types
import os
import sys
from typing import Tuple, Iterable, Optional, Set, List, Dict, Any, Union
from getpass import getuser
from pwd import getpwuid
import subprocess

from powar.global_config import GlobalConfig
from powar.settings import AppSettings
from powar.util import saved_sys_properties, render_template, realpath, read_header, run_command, UserError

logger: logging.Logger = logging.getLogger(__name__)


class ModuleConfigApi:
    opts: Dict[Any, Any]
    local: Dict[Any, Any]

    _man: 'ModuleConfigManager'

    def __init__(self, man: 'ModuleConfigManager', opts: Dict[Any, Any],
                 local: Dict[Any, Any]):
        self.opts = opts
        self.local = local
        self._man = man

    def install(
        self,
        entries: Union[Set[Tuple[str, str]], Dict[str, str]],
    ) -> None:
        '''
        Install files
        '''
        if isinstance(entries, set):
            self._man.install_entries(entries)
        elif isinstance(entries, dict):
            self._man.install_entries(entries.items())
        else:
            raise TypeError(
                f"invalid argument type to install: {type(entries)}")

    def install_bin(
        self,
        entries: Union[Set[Tuple[str, str]], Dict[str, str]],
    ) -> None:
        '''
        Install files
        '''
        if isinstance(entries, set):
            self._man.install_entries(entries, binary=True)
        elif isinstance(entries, dict):
            self._man.install_entries(entries.items(), binary=True)
        else:
            raise TypeError(
                f"invalid argument type to install_bin: {type(entries)}")

    def execute(
        self,
        command: str,
        stdin=Optional[str],
        return_stdout=False,
        decode_stdout=True,
        wait=True,
    ) -> Optional[Union[str, bytes]]:
        '''
        Run command and return stdout if any
        '''
        return self._man.execute_command(command, stdin, return_stdout,
                                         decode_stdout, wait)

    def render(self, x: str) -> str:
        '''
        Render jinja templated string and return it
        '''
        return self._man.render_template(x)


class ModuleConfigManager:
    _directory: str
    _settings: AppSettings
    _global_config: GlobalConfig
    _api: ModuleConfigApi

    _current_user: str = getuser()

    _opts: Dict[Any, Any]
    _local: Dict[Any, Any] = {}

    _module_name: str
    _config_path: str
    _header: Optional[Dict[Any, Any]] = None

    def __init__(
        self,
        directory: str,
        global_config: GlobalConfig,
        app_settings: AppSettings,
    ):
        self._directory = directory
        self._global_config = global_config
        self._settings = app_settings

        self._opts = global_config.opts

        self._module_name = os.path.basename(self._directory)
        self._config_path = os.path.join(self._directory,
                                         app_settings.module_config_filename)

    def run(self) -> None:
        self._ensure_depends_are_met()

        api = ModuleConfigApi(self, self._opts, self._local)

        module = types.ModuleType('powar')
        module.p = api  # type: ignore
        module.__file__ = self._config_path

        with open(self._config_path, 'rb') as f:
            source = f.read()

        code = compile(source, self._config_path, 'exec')

        # Save and restore sys variables and cwd
        old_cwd = os.getcwd()
        with saved_sys_properties():
            if self._directory not in sys.path:
                sys.path.insert(0, self._directory)
            os.chdir(self._directory)
            exec(code, module.__dict__)
            os.chdir(old_cwd)

    def get_system_packages(self) -> List[str]:
        header = self._read_header()
        return header.get('system_packages', [])

    def install_entries(self,
                        entries: Iterable[Tuple[str, str]],
                        binary=False) -> None:
        dir_files = os.listdir(self._directory)

        if binary:
            for src, dest in entries:
                self._install_bin(src, dest)
        else:
            for src, dest in entries:
                with open(os.path.join(self._directory, src), 'r') as f:
                    src_contents = f.read()
                rendered = self.render_template(src_contents)
                self._install_file(src, dest, content=rendered)

    def execute_command(self, command: str, stdin: Optional[str],
                        return_stdout: bool, decode_stdout: bool,
                        wait: bool) -> Optional[Union[str, bytes]]:
        stdout = None
        if not self._settings.dry_run:
            stdout = run_command(command, self._directory, return_stdout,
                                 decode_stdout, wait)
        logger.info(
            f"Ran{'' if wait else ' (in bg)'}: {command} for {self._config_path}"
        )
        return stdout

    def _ensure_depends_are_met(self) -> None:
        header = self._read_header()

        depends = header.get('depends')
        if not depends:
            return

        if self._module_name in depends:
            raise UserError(
                f"module \"{self._module_name}\" cannot depend on itself")

        missing = depends - set(self._global_config.modules)

        if missing:
            raise UserError(*(
                f"module \"{self._module_name}\" depends on \"{module}\", " \
                f"but this is not enabled" for module in missing
            ))

    def render_template(
        self,
        contents: str,
    ) -> str:
        return render_template(
            contents,
            variables={
                'local': self._local,
                **self._opts,
            },
            directory=self._directory,
        )

    def _read_header(self) -> Dict[Any, Any]:
        if self._header is None:
            self._header = read_header(self._config_path)
        return self._header

    def _can_install_without_root(self, dest: str) -> bool:
        try:
            owner_of_dest = getpwuid(os.stat(dest).st_uid).pw_name
        except FileNotFoundError:
            owner_of_dest = getpwuid(os.stat(
                os.path.dirname(dest)).st_uid).pw_name

        if not owner_of_dest == self._current_user:
            return False
        return True

    def _install_file(self, src: str, dest: str, content: str) -> None:
        dest = realpath(dest)

        command = ['tee', dest]

        if not self._can_install_without_root(dest):
            if not self._settings.switch_to_root:
                logger.warn(
                    f"installing at \"{dest}\" requires to be in root mode, skipping"
                )
                return
            command = ['sudo', '-E', *command]

        if not self._settings.dry_run:
            run_command(' '.join(command),
                        self._directory,
                        stdin=str.encode(content + '\n'))
        logger.info(f"Done: {src} -> {dest}")

    def _install_bin(self, src: str, dest: str) -> None:
        dest = realpath(dest)

        command = ['cp', src, dest]

        if not self._can_install_without_root(dest):
            if not self._settings.switch_to_root:
                logger.warn(
                    f"installing at \"{dest}\" requires to be in root mode, skipping"
                )
                return
            command = ['sudo', '-E', *command]

        if not self._settings.dry_run:
            run_command(' '.join(command), self._directory)
        logger.info(f"Done (bin): {src} -> {dest}")



import logging
import types
import os
import sys
from typing import Iterator, Optional, Set, List, Dict, Any, Union
from getpass import getuser

from powar.settings import AppSettings
from powar.util import saved_sys_properties, read_header, run_command

logger: logging.Logger = logging.getLogger(__name__)


class GlobalConfig:
    modules: List[str]
    opts: Dict[Any, Any] = {}


class GlobalConfigApi:
    opts: Dict[Any, Any]

    _man: 'GlobalConfigManager'

    def __init__(self, man: 'GlobalConfigManager', opts: Dict[Any, Any]):
        self.opts = opts
        self._man = man

    def execute(
        self,
        command: str,
        stdin=Optional[str],
        return_stdout=False,
        decode_stdout=True,
        wait=True,
    ) -> Optional[Union[str, bytes]]:
        '''
        Run command and return stdout if any
        '''
        return self._man.execute_command(command, stdin, return_stdout,
                                         decode_stdout, wait)


class GlobalConfigManager:
    _directory: str
    _settings: AppSettings
    _api: GlobalConfigApi

    _global_config: GlobalConfig = GlobalConfig()

    _config_path: str
    _header: Optional[Dict[Any, Any]] = None

    def __init__(
        self,
        directory: str,
        app_settings: AppSettings,
    ):
        self._directory = directory
        self._settings = app_settings

        self._config_path = os.path.join(self._directory,
                                         app_settings.global_config_filename)

    def get_global_config(self) -> GlobalConfig:
        api = GlobalConfigApi(self, self._global_config.opts)

        module = types.ModuleType('powar')
        module.p = api  # type: ignore
        module.__file__ = self._config_path

        with open(self._config_path, 'rb') as f:
            source = f.read()

        code = compile(source, self._config_path, 'exec')

        # Save and restore sys variables and cwd
        old_cwd = os.getcwd()
        with saved_sys_properties():
            if self._directory not in sys.path:
                sys.path.insert(0, self._directory)
            os.chdir(self._directory)
            exec(code, module.__dict__)
            os.chdir(old_cwd)

        self._global_config.modules = self._get_modules()
        return self._global_config

    def execute_command(self, command: str, stdin: Optional[str],
                        return_stdout: bool, decode_stdout: bool,
                        wait: bool) -> Optional[Union[str, bytes]]:
        stdout = None
        if not self._settings.dry_run:
            stdout = run_command(command, self._directory, return_stdout,
                                 decode_stdout, wait)
        logger.info(f"Ran: {command} for {self._config_path}")
        return stdout

    def _read_header(self) -> Dict[Any, Any]:
        if self._header is None:
            self._header = read_header(self._config_path)
        return self._header

    def _get_modules(self) -> List[str]:
        header = self._read_header()
        return header.get('modules', [])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="powar",
    version="1.0.2",
    author="Constantine Theocharis",
    author_email="kontheocharis@gmail.com",
    description="A configuration manager unlike any other!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kontheocharis/powar",
    packages=setuptools.find_packages(),

    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
    ],
    python_requires='>=3.6',
    install_requires=open('requirements.txt').read().splitlines(),
    entry_points={
        'console_scripts': [
            'powar=powar.main:main',
        ],
    },
)



import argparse
import os
import logging
import time
import shutil
import sys
import subprocess
from typing import cast, Iterable

from powar.module_config import ModuleConfigManager
from powar.global_config import GlobalConfigManager, GlobalConfig
from powar.settings import AppSettings, AppMode, AppLogLevel
from powar.util import realpath, UserError

LOGGING_FORMAT = "%(levelname)s: %(message)s"
logger: logging.Logger

ROOT_FLAGS = ("--root", "-r")


def parse_args_into(app_settings: AppSettings) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--dry-run",
        dest="dry_run",
        help="don't modify any files, just show what would be done",
        action="store_true")

    parser.add_argument("--template-dir",
                        dest="template_dir",
                        help="use a custom directory for templates")

    parser.add_argument("--config-dir",
                        dest="config_dir",
                        help="use a custom directory for configuration")

    parser.add_argument(
        *ROOT_FLAGS,
        dest="switch_to_root",
        action="store_true",
        help=
        "run powar in sudo mode to be able to install files in places outside $HOME")

    group = parser.add_mutually_exclusive_group()
    group.add_argument("-q",
                       "--quiet",
                       help="supress output",
                       action="store_const",
                       dest="log_level",
                       const=AppLogLevel.QUIET)
    group.add_argument("-v",
                       "--verbose",
                       help="be verbose",
                       action="store_const",
                       dest="log_level",
                       const=AppLogLevel.VERBOSE)

    subparsers = parser.add_subparsers(help="mode to use",
                                       dest="mode command")

    # Install mode
    parser_install = subparsers.add_parser(
        "install",
        help="install specified modules (empty argument installs all modules)")
    parser_install.set_defaults(mode=AppMode.INSTALL)
    parser_install.add_argument(
        "modules_to_consider",
        nargs="*",
        metavar="MODULE",
        help="module(s) to install (empty argument installs all modules)")

    # List-packages mode
    parser_list = subparsers.add_parser(
        "list", help="list system packages required to install config files")
    parser_list.set_defaults(mode=AppMode.LIST_PACKAGES)
    parser_list.add_argument(
        "modules_to_consider",
        nargs="*",
        metavar="MODULE",
        help=
        "module(s) to list packages for (empty argument lists for all modules)"
    )

    # New module mode
    parser_new = subparsers.add_parser("new", help="create a new powar module")
    parser_new.set_defaults(mode=AppMode.NEW_MODULE)
    parser_new.add_argument("new_module_name",
                            metavar="MODULE_NAME",
                            help="name of the new module to be created")

    # Init mode
    parser_init = subparsers.add_parser(
        "init", help="create the folders required for powar")
    parser_init.set_defaults(mode=AppMode.INIT)

    parser.parse_args(namespace=app_settings)
    return parser


def run_init(app_settings: AppSettings) -> None:
    os.makedirs(app_settings.config_dir, exist_ok=True)
    global_config_path = os.path.join(app_settings.config_dir,
                                      app_settings.global_config_filename)

    if not os.path.exists(global_config_path):
        shutil.copy(
            os.path.join(app_settings.data_path,
                         app_settings.global_config_template_filename),
            global_config_path)
        print(f"{global_config_path} created.")
    else:
        logger.warn(f"{global_config_path} exists, skipping.")

    try:
        os.makedirs(app_settings.template_dir)
        print(f"{app_settings.template_dir}/ created.")
    except FileExistsError:
        logger.warn(f"{app_settings.template_dir} exists, skipping.")


def run_new_module(app_settings: AppSettings) -> None:
    if not os.path.exists(app_settings.template_dir):
        raise UserError(f"{app_settings.template_dir} doesn't exist.")

    module_dir = os.path.join(app_settings.template_dir,
                              cast(str, app_settings.new_module_name))
    try:
        os.makedirs(module_dir)
    except FileExistsError:
        raise UserError(f"{module_dir} already exists.")

    module_config_path = os.path.join(module_dir,
                                      app_settings.module_config_filename)
    shutil.copy(
        os.path.join(app_settings.data_path,
                     app_settings.module_config_template_filename),
        module_config_path)
    print(f"{module_config_path} created.")


def run_list_packages(app_settings: AppSettings,
                      module_directories: Iterable[str],
                      global_config: GlobalConfig) -> None:
    for directory in module_directories:
        manager = ModuleConfigManager(directory, global_config, app_settings)
        for package in manager.get_system_packages():
            print(f"{os.path.basename(directory)}:")
            print(package)
            print('\n')


def run_install(app_settings: AppSettings, module_directories: Iterable[str],
                global_config: GlobalConfig) -> None:
    for directory in module_directories:
        manager = ModuleConfigManager(directory, global_config, app_settings)
        manager.run()


def main() -> None:
    app_settings = AppSettings()
    parser = parse_args_into(app_settings)

    # set logging level from arguments
    logging.basicConfig(level=app_settings.log_level.into_logging_level(),
                        format=LOGGING_FORMAT)
    logger = logging.getLogger(__name__)

    # resolve $VARIABLES and ~, ensure absolute
    dirs_to_resolve = ("template_dir", "config_dir", "cache_dir")
    for var in dirs_to_resolve:
        app_settings[var] = realpath(app_settings[var])
        if not os.path.isabs(app_settings[var]):
            parser.error(f"{var} needs to be absolute")

    try:
        if app_settings.mode == AppMode.INIT:
            return run_init(app_settings)

        if app_settings.mode == AppMode.NEW_MODULE:
            return run_new_module(app_settings)

        # cache_man = CacheManager(app_settings.cache_dir)

        global_config = GlobalConfigManager(app_settings.config_dir,
                                            app_settings).get_global_config()

        directories = [
            os.path.join(app_settings.template_dir, module)
            for module in global_config.modules
        ]

        if app_settings.mode == AppMode.LIST_PACKAGES:
            return run_list_packages(app_settings, directories, global_config)

        # Main logic
        if app_settings.mode == AppMode.INSTALL:
            if not directories:
                return logger.info("No files to install, exiting.")

            return run_install(app_settings, directories, global_config)

    except UserError as error:
        for arg in error.args:
            logger.error(arg)
    return None
